(* Verschlsseln von Dateien
    J. Rathlev, D-24222 Schwentinental

 -----------------------------------------------------------------------------
   Cipher routines from
     Delphi Encryption Compendium Part I (Version 5.2)
     2006-2008 Hagen Reddmann, Arvid Winkelsdorf
 -----------------------------------------------------------------------------

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Oct. 2005
         2 - May 2015
   last modified: Nov. 2023
   *)

unit CipherMain;

interface

uses
  WinApi.Windows, WinApi.Messages, System.SysUtils, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.Buttons, Vcl.WinXCtrls,
  Vcl.ComCtrls, Vcl.ExtCtrls, System.IniFiles, HListBox, WinShell,
  System.ImageList, Vcl.ImgList;

resourcestring
  Prog = 'File encryption';
  Vers = ' - Version 3.0';

const
  BSize = 1024*1024;
  ThPr = tpNormal;

  // command line options
  coComp   = 'compress';
  coDec    = 'decrypt';
  coEnc    = 'encrypt';
  coInfo   = 'info';
  coMethod = 'methode';
  coDest   = 'output';
  coPwd    = 'password';
  coRecur  = 'recurse';
  coSource = 'source';
  coWait   = 'wait';

  cmAes   = 'aes';
  cmBlow  = 'blowfish';
  cm3Des  = '3des';
  cmTwo   = 'twofish';

{$I ..\..\Common\JrGlobal.pas}

type
  THauptForm = class(TForm)
    Label1: TLabel;
    Label2: TLabel;
    EndeBtn: TBitBtn;
    OpenDialog: TOpenDialog;
    Label3: TLabel;
    DestDirEdit: THistoryCombo;
    DestSelectBtn: TSpeedButton;
    SourceDirEdit: THistoryCombo;
    FileEdit: THistoryCombo;
    SourceSelectBtn: TSpeedButton;
    FileBtn: TSpeedButton;
    StatusBar: TStaticText;
    cbSubDir: TCheckBox;
    rgpDirection: TRadioGroup;
    laMethod: TLabel;
    cbxCipher: TComboBox;
    cbComp: TCheckBox;
    cbInfo: TCheckBox;
    aiCipher: TActivityIndicator;
    ImageList: TImageList;
    InfoBtn: TBitBtn;
    btnProcess: TBitBtn;
    procedure FormCreate(Sender: TObject);
    procedure EndeBtnClick(Sender: TObject);
    procedure InfoBtnClick(Sender: TObject);
    procedure DestSelectBtnClick(Sender: TObject);
    procedure FileBtnClick(Sender: TObject);
    procedure btnProcessClick(Sender: TObject);
    procedure SourceSelectBtnClick(Sender: TObject);
    procedure rgpDirectionClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure SourceDirEditCloseUp(Sender: TObject);
    procedure FormActivate(Sender: TObject);
  private
    { Private-Deklarationen }
    ArbPfad,
    PrgPfad,PrgName,
    IniName,LSrc,LDest  : string;
    Pwd                 : AnsiString;
    CmdOptions,
    InstantRun,Prompt,
    Processing,Stopped  : boolean;
  public
    { Public-Deklarationen }
  end;

var
  HauptForm: THauptForm;

implementation

{$R *.DFM}
{$R *.RES}

uses WinApi.ShlObj, ShellDirDlg, WinApiUtils, PathUtils, UnitConsts, StringUtils,
  WinUtils, CipherTools, NewPwdDlg, GnuGetText, FileErrors, MsgDialogs;

const
  IniExt = 'ini';

  (* INI-Sektionen *)
  CfGSekt = 'Config';
  SourceSekt = 'Source';
  DestSekt = 'Dest';
  FileSekt = 'Files';

  (* INI-Variablen *)
  iniTop = 'Top';
  iniLeft = 'Left';
  iniDir = 'Direction';
  iniSubDir = 'Subdirectories';
  iniComp   = 'Compress';
  iniCipher = 'Algorithm';
  iniESource = 'EnSourceDir';
  iniEDest   = 'EnDestDir';
  iniDSource = 'DeSourceDir';
  iniDDest   = 'DeDestDir';

{ ------------------------------------------------------------------- }
(* INI-File lesen *)
procedure THauptForm.FormCreate(Sender: TObject);
var
  IniFile  : TIniFile;
  Files    : string;

  procedure SetOption (s : string);
  var
    j   : integer;
  begin
    if length(s)>0 then begin
      if (s[1]='/') or (s[1]='-') then begin
        delete (s,1,1);
        if ReadOptionValue(s,coMethod) then begin
          if CompareOption(s,cmAes) then cbxCipher.ItemIndex:=0
          else if CompareOption(s,cmBlow) then cbxCipher.ItemIndex:=1
          else if CompareOption(s,cm3Des) then cbxCipher.ItemIndex:=2
          else if CompareOption(s,cmTwo) then cbxCipher.ItemIndex:=3;
          end
        else if ReadOptionValue(s,coSource) then SourceDirEdit.Text:=s
        else if ReadOptionValue(s,coDest) then DestDirEdit.Text:=s
        else if ReadOptionValue(s,coPwd) then Pwd:=s
        else if CompareOption(s,coEnc) then InstantRun:=true
        else if CompareOption(s,coDec) then begin
          InstantRun:=true; rgpDirection.ItemIndex:=1;
          end
        else if CompareOption(s,coWait) then Prompt:=true
        else if CompareOption(s,coComp) then cbComp.Checked:=true
        else if CompareOption(s,coRecur) then cbSubDir.Checked:=true
        else if CompareOption(s,coInfo) then cbInfo.Checked:=true;
        end
      else if length(Files)=0 then Files:=MakeQuotedStr(s,[Space,';'])
        else Files:=Files+','+MakeQuotedStr(s,[Space,';']);
      end;
    end;

  procedure ReadOptions;
  var
    s : string;
    i : integer;
  begin
    Files:='';
    if ParamCount>0 then begin
      CmdOptions:=true;
      rgpDirection.ItemIndex:=0;
      SourceDirEdit.Text:=GetCurrentDir; DestDirEdit.Text:=GetCurrentDir;
      cbComp.Checked:=false; cbSubDir.Checked:=false; cbInfo.Checked:=false;
      for i:=1 to ParamCount do begin
        s:=ParamStr(i);
        if (s[1]='@') then begin  // Steuerdatei auswerten
          delete (s,1,1);
          s:=ReadOptionFile(s);
          repeat
            SetOption(ReadNxtStr(s,'|'));
            until length(s)=0;
          end
        else SetOption(s);
        end;
      end
    else CmdOptions:=false;
    end;

begin
  TranslateComponent (self);
  PrgPfad:=ExtractFilePath(Application.ExeName);
  PrgName:=ExtractFileName(DelExt(Application.ExeName));
  ArbPfad:=GetDesktopFolder(CSIDL_APPDATA);
  cbxCipher.Items.CommaText:=CipherAlgos;
  IniName:=Erweiter(ArbPfad,PrgName,IniExt);
  SourceDirEdit.LoadFromIni (IniName,SourceSekt);
  DestDirEdit.LoadFromIni (IniName,DestSekt);
  IniFile:=TIniFile.Create(IniName);
  with IniFile do begin
    Top:=ReadInteger(CfgSekt,iniTop,Top);
    Left:=ReadInteger(CfgSekt,iniLeft,Left);
    rgpDirection.ItemIndex:=ReadInteger(CfgSekt,iniDir,0);
    cbSubDir.Checked:=ReadBool(CfgSekt,iniSubDir,false);
    cbComp.Checked:=ReadBool(CfgSekt,iniComp,false);
    cbxCipher.ItemIndex:=ReadInteger(CfgSekt,iniCipher,0);
    if rgpDirection.ItemIndex=0 then begin
      SourceDirEdit.Text:=ReadString(CfgSekt,iniESource,'');
      DestDirEdit.Text:=ReadString(CfgSekt,iniEDest,'');
      LSrc:=ReadString(CfgSekt,iniDSource,'');
      LDest:=ReadString(CfgSekt,iniDDest,'');
      end
    else begin
      SourceDirEdit.Text:=ReadString(CfgSekt,iniDSource,'');
      DestDirEdit.Text:=ReadString(CfgSekt,iniDDest,'');
      LSrc:=ReadString(CfgSekt,iniESource,'');
      LDest:=ReadString(CfgSekt,iniEDest,'');
      end;
    Free;
    end;
  Pwd:='';
  InstantRun:=false; Prompt:=false;
  ReadOptions;
  if length(Files)=0 then begin
    with SourceDirEdit do if length(Text)=0 then ItemIndex:=0;
    with DestDirEdit do if length(Text)=0 then ItemIndex:=0;
    with FileEdit do begin
      LoadFromIni (IniName,FileSekt);
      ItemIndex:=0;
      end;
    end
  else with FileEdit do Text:=Files;
  Processing:=false; Stopped:=false;
  end;

procedure THauptForm.FormDestroy(Sender: TObject);
var
  IniFile  : TIniFile;
begin
  IniFile:=TIniFile.Create(IniName);
  if not CmdOptions then with IniFile do begin
    WriteInteger(CfgSekt,iniTop,Top);
    WriteInteger(CfgSekt,iniLeft,Left);
    WriteInteger(CfgSekt,iniDir,rgpDirection.ItemIndex);
    WriteBool(CfgSekt,iniSubDir,cbSubDir.Checked);
    WriteBool(CfgSekt,iniComp,cbComp.Checked);
    WriteInteger(CfgSekt,iniCipher,cbxCipher.ItemIndex);
    if rgpDirection.ItemIndex=0 then begin
      WriteString(CfgSekt,iniESource,SourceDirEdit.Text);
      WriteString(CfgSekt,iniEDest,DestDirEdit.Text);
      WriteString(CfgSekt,iniDSource,LSrc);
      WriteString(CfgSekt,iniDDest,LDest);
      end
    else begin
      WriteString(CfgSekt,iniDSource,SourceDirEdit.Text);
      WriteString(CfgSekt,iniDDest,DestDirEdit.Text);
      WriteString(CfgSekt,iniESource,LSrc);
      WriteString(CfgSekt,iniEDest,LDest);
      end;
    Free;
    end;
  end;

procedure THauptForm.FormShow(Sender: TObject);
begin
//  rgpDirectionClick(Sender);
  end;

procedure THauptForm.FormActivate(Sender: TObject);
begin
  if InstantRun then begin
    btnProcessClick(Sender);
    if not Prompt then Close;
    end;
  end;

{ ------------------------------------------------------------------- }
(* Programmende *)
procedure THauptForm.EndeBtnClick(Sender: TObject);
begin
  Close;
  end;

{ ------------------------------------------------------------------- }
procedure THauptForm.InfoBtnClick(Sender: TObject);
begin
  InfoDialog(GetFileVersionName(Application.Exename,Prog,Vers)+sLineBreak+CopRgt+sLineBreak+
                         _('E-Mail: ')+EMailAdr+sLineBreak+
                         _('Cipher routines from:')+sLineBreak+
                         _('Delphi Encryption Compendium - Version 5.2')+sLineBreak+
                         '( '+_('2006-2008 Hagen Reddmann, Arvid Winkelsdorf')+')');
  end;

{ ------------------------------------------------------------------- }
procedure THauptForm.rgpDirectionClick(Sender: TObject);
var
  s : string;
begin
  if rgpDirection.ItemIndex=0 then begin
    cbxCipher.Enabled:=true;
    cbComp.Enabled:=true;
    cbInfo.Enabled:=true;
    cbxCipher.Visible:=true;
    laMethod.Visible:=true;
    end
  else begin
    cbxCipher.Enabled:=false;
    cbComp.Enabled:=false;
    cbInfo.Enabled:=false;
    cbxCipher.Visible:=false;
    laMethod.Visible:=false;
    end;
  if not Visible then begin  // read command line
    Exit;
    end
  else begin
    if rgpDirection.ItemIndex=0 then begin
      FileEdit.Text:='*.*';
      end
    else begin
      FileEdit.Text:='*'+CryptExt;
      end;
    Pwd:='';
    if length(LSrc)>0 then begin
      s:=SourceDirEdit.Text; SourceDirEdit.Text:=LSrc; LSrc:=s;
      end
    else LSrc:=SourceDirEdit.Text;
    if length(LDest)>0 then begin
      s:=DestDirEdit.Text; DestDirEdit.Text:=LDest; LDest:=s;
      end
    else LDest:=DestDirEdit.Text;
    end;
  end;

procedure THauptForm.SourceSelectBtnClick(Sender: TObject);
var
  s,t : string;
begin
  s:=SourceDirEdit.Text;
  if length(s)=0 then s:=GetDesktopFolder(CSIDL_PERSONAL);
  if rgpDirection.ItemIndex=0 then t:=_('Directory with files to be encrypted')
  else t:=_('Directory with files to be decrypted');
  if ShellDirDialog.Execute (t,true,true,true,'',s) then begin
    with SourceDirEdit do begin
      Text:=s; AddItem(s);
      end;
    with DestDirEdit do if length(Text)=0 then begin
      Text:=s; AddItem(s);
      end;
    if rgpDirection.ItemIndex=0 then FileEdit.Text:='*.*'
    else FileEdit.Text:='*'+CryptExt;
    end;
  end;

procedure THauptForm.DestSelectBtnClick(Sender: TObject);
var
  s,t : string;
begin
  s:=DestDirEdit.Text;
  if length(s)=0 then s:=SourceDirEdit.Text;
  if length(s)=0 then s:=GetDesktopFolder(CSIDL_PERSONAL);
  if rgpDirection.ItemIndex=0 then t:=_('Directory for encrypted files')
  else t:=_('Directory for decrypted files');
  if ShellDirDialog.Execute (t,true,true,true,'',s) then begin
    with DestDirEdit do begin
      Text:=s; AddItem(s);
      end;
    end;
  end;

procedure THauptForm.FileBtnClick(Sender: TObject);
var
  i : integer;
  s : string;
begin
  with OpenDialog do begin
    InitialDir:=SourceDirEdit.Text;
    Filename:='';
    if rgpDirection.ItemIndex=0 then begin
      Filter:=_('all|*.*');
      Title:=_('Select files to be encrypted');
      end
    else begin
      Filter:=_('Encrypted files|*')+CryptExt+_('|all|*.*');
      Title:=_('Select files to be decrypted');
      end;
    if Execute then begin
      cbSubDir.Checked:=false;
      s:=ExtractFilePath(Files.Strings[0]);
      with SourceDirEdit do if s<>Text then begin
        Text:=s; AddItem(s);
        end;
      with Files do begin
        s:='';
        for i:=0 to Count-1 do s:=s+ExtractFilename(Strings[i])+';';
        end;
      delete (s,length(s),1);
      with FileEdit do begin
        Text:=s; AddItem(s);
        end;
      end;
    end;
  end;

procedure THauptForm.btnProcessClick(Sender: TObject);
var
  FListe        : TStringList;
  method,
  erc,fcnt,i,n  : integer;
  s,t,dp,
  sName,
  fname,oname   : string;
  encrypt       : boolean;

  procedure ScanDir (Dir,Mask : string);
  var
    DirInfo    : TSearchRec;
    Findresult : integer;
  begin
    if cbSubDir.Checked then begin
      FindResult:=FindFirst (Erweiter(Dir,'*','*'),faAnyFile,DirInfo);
      while FindResult=0 do with DirInfo do begin
        if NotSpecialDir(Name) then begin
          if((Attr and faDirectory)<>0) then ScanDir (Erweiter(Dir,Name,''),Mask);
          end;
        FindResult:=FindNext (DirInfo);
        end;
      FindClose(DirInfo);
      end;
    FindResult:=FindFirst (Erweiter(Dir,Mask,''),faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) then FListe.Add(Erweiter(Dir,Name,''));
      FindResult:=FindNext (DirInfo);
      end;
    FindClose(DirInfo);
    end;

begin
  if Processing then Stopped:=true
  else begin
    t:=FileEdit.Text;
    encrypt:=rgpDirection.ItemIndex=0;
//    Pwd:='abcd1234'; //***
    if (length(t)>0) and ((length(Pwd)>0) or NewPwdDialog.Execute(BottomRightPos(btnProcess),
                        _('Encryption password'),Pwd,encrypt,encrypt,8)) then begin
      Processing:=true; erc:=0;
      aiCipher.Animate:=true;
      with StatusBar do begin
        Font.Color:=clBlack;
        Caption:='';
        end;
      FListe:=TStringList.Create;
      fcnt:=0;
      with btnProcess do begin
//        Glyph.LoadFromResourceName(HInstance,'STOP');
        Glyph:=nil;
        ImageList.GetBitmap(1,Glyph);
        Caption:=_('Stop');
        Hint:=_('Cancel processing');
        end;
      repeat
        s:=ReadNxtStr (t,';');
        FListe.Clear;
        if ContainsFullPath(s) then FListe.Add(s)
        else ScanDir (SourceDirEdit.Text,s);
        until (length(t)=0);
      with FListe do if Count>0 then for i:=0 to Count-1 do begin
        if Stopped then Break;
        inc(fcnt);
        fName:=Strings[i];
        sName:=ExtractRelativePath(SetDirName(SourceDirEdit.Text),Strings[i]);
        if ContainsFullPath(s) then oName:=sName
        else oName:=SetDirName(DestDirEdit.Text)+sName;
        dp:=ExtractFilePath(oName);
        if not DirectoryExists(dp) then ForceDirectories(dp);
        if rgpDirection.ItemIndex=0 then begin
          StatusBar.Caption:=_('Encrypting: ')+sName;
          oName:=oName+CryptExt;
          method:=cbxCipher.ItemIndex; //or cmFileInfo;
          if cbComp.Checked then method:=method or cmCompress;
          if cbInfo.Checked then method:=method or cmFileInfo;
          with TEncryptThread.Create (fName,oName,Pwd,method,false,BSize,ThPr) do begin
            repeat
              Sleep(100);
              Application.ProcessMessages;
              until Done;
            erc:=ErrorCode;
            Free;
            end;
          end
        else begin
          oName:=DelExt(oName);
          with TDecryptThread.Create (fName,oName,Pwd,false,BSize,ThPr) do begin
            Sleep(100);
            n:=CipherAlgo;
            if n>=0 then s:=cbxCipher.Items[n] else s:='';
            StatusBar.Caption:=_('Decrypting')+' ('+s+'): '+sName;
            repeat
              Sleep(100);
              Application.ProcessMessages;
              until Done;
            erc:=ErrorCode;
            Free;
            end;
          end;
        if erc<>0 then Break;
        end;
      FListe.Free;
      with StatusBar do begin
        if (erc=0) and (fcnt>0) then begin
          Font.Color:=clBlue;
          if rgpDirection.ItemIndex=0 then Caption:=' '+IntToStr(fcnt)+_(' file(s) encrypted')
          else Caption:=' '+IntToStr(fcnt)+_(' file(s) decrypted (')+s+')';
          end
        else begin
          Font.Color:=clRed;
          if erc>0 then Caption:=GetCopyErrMsg(erc)+': '+sName
          else Caption:=_('No matching files found');
          end;
        end;
      aiCipher.Animate:=false;
      with btnProcess do begin
//        Glyph.LoadFromResourceName(HInstance,'START');
        Glyph:=nil;
        ImageList.GetBitmap(0,Glyph);
        Caption:=_('Start');
        Hint:=_('Start processing');
        end;
      Processing:=false; Stopped:=false;
      end
    else StatusBar.Caption:=_('Nothing to do!');
    end;
  end;

procedure THauptForm.SourceDirEditCloseUp(Sender: TObject);
begin
  if rgpDirection.ItemIndex=0 then FileEdit.Text:='*.*'
  else FileEdit.Text:='*'+CryptExt;
  end;

end.
